% SorensenMyken92
%
% This file reconstructs the results in Section 4.1 of:
%
% Sorensen and Myken, Unsteady actuator disc model for horizontal axis
% wind turbines, Journal of Wind Engineering and Industrial Aerodynamics,
% vol. 39, 1992, p. 139-149.
%
% The paper describes a method to solve for the flow around an actuator
% disk assuming incompressible, inviscid, and axis-symmetric flow.

%% Computational Domain
clear all
close all
clc

R = 1;                      % Actuator disk radius, m
Ad = pi*R^2;                % Actuator disk area, m^2

Nx = 120;                   % # of points in x direction
xmax = 6*R;                 % Maximum x value, m
x = linspace(0,xmax,Nx)';   % Gridpoints in x direction, m
dx = x(2)-x(1);             % Grid spacing in x direction, m

Nr = 40;                    % # of points in r direction
rmax = 2*R;                 % Maximum r value, m
r = linspace(0,rmax,Nr)';   % Gridpoints in r direction, m
dr = r(2)-r(1);             % Grid spacing in r direction, m
rmat = repmat(r',[Nx,1]);   % Matrix of radius values at all (i,j), m

Nt = 100;                    % # of time points
tmax = 5;                   % Maximum time value, sec
t = linspace(0,tmax,Nt)';   % Gridpoints in time, sec
dt = t(2)-t(1);             % Time steps, sec

%% Model Parameters
rho = 1.225;                % Density of air, kg/m^3

%% Initial Condition
U0 = 8;                    % Inlet streamwise velocity, m/s

u = repmat(U0,[Nx Nr]);     % Initial velocity in x direction, m/s
v = zeros(Nx,Nr);           % Initial velocity in r direction, m/s
%w = zeros(Nx,Nr);           % Initial velocity in theta direction, m/s

psi = repmat(-0.5*r.^2*U0,[1,Nx])';     % Stream function, m^3/sec
om = zeros(Nx,Nr);                      % Azimuthal vorticity, rad/sec

%% Actuator Disk Location and Force
% fx is the force per unit volume, N/m^3
% dfx is d(fx)/dr, N/m^4
xd = 2*R;                   % Location of rotor in x direction, m
[~,id] = min( abs(x-xd) );  % Gridpoint x-index of rotor location
jd = max(find(r<=R));       % Gridpoint r-index of rotor radius

CT = 0.68;                   % Thrust coefficient, unitless
fx = zeros(Nr,1);
fx(1:jd) = (0.75*rho*U0^2*CT/dx)*sqrt( 1-r(1:jd).^2/R^2 );
dfx = zeros(Nr,1);
dfx(1:jd) = (0.75*rho*U0^2*CT/dx)*(-r(1:jd)/R^2 )./sqrt(1-r(1:jd).^2/R^2);
%dfx2 = [diff(fx)/dr; 0];

fx = -fx;
dfx = -dfx;

%% Poisson Matrix
% Form matrices used in discretized Poisson equation for streamfunction.
% The LU factorization is computed to speed up solution.
[Mi,Mbc,Bbc] = CylPoissonMatrix(x,r,U0);
Mp = [Mi; Mbc];
[Lp,Up]=lu(Mp);

%%  Numerical Solution
for n=1:Nt
    % Alternating Direction Implicit: Update vorticity
    
    % Step A: Implicit in x and Explicit in r
    vom = v.*om;
    for j=2:(Nr-1)
        % Grab voriticity and velocities at radius rj
        omj = om(:,j);
        uj = u(:,j);
        vj = v(:,j);
        
        % Form right side of algebraic equation
        bet = omj;
        bet(id) = bet(id)-(dt/2/rho)*dfx(j);
        pidx = find(vj>=0);
        bet(pidx) = bet(pidx)-(dt/2/dr)*(vom(pidx,j)-vom(pidx,j-1));
        nidx = find(vj<0);
        bet(nidx) = bet(nidx)-(dt/2/dr)*(vom(nidx,j+1)-vom(nidx,j));
        bet([1 end]) = 0;
        
        % Form tridiagonal matrix of coefficients based on upwinding
        c = -(dt/2/dx)*( uj(2:end-1)>=0 ).*uj(1:end-2);
        c = [c; -1; 0];
        b = 1+(dt/2/dx)*( uj(2:end-1)>=0 ).*uj(2:end-1);
        b = b -  (dt/2/dx)*( uj(2:end-1)<0 ).*uj(2:end-1);
        b = [1; b; 1];
        a = (dt/2/dx)*( uj(2:end-1)<0 ).*uj(3:end);
        a = [0; 0; a];
        M = spdiags([c b a],[-1 0 1],Nx,Nx);
        
        % Solve equations for updated vorticity
        om(:,j) = M\bet;
    end
    
    % Step B: Implicit in r and Explicit in x
    uom = u.*om;
    for i=2:(Nx-1)
        % Grab voriticity and velocities at location xi
        omi = om(i,:);
        ui = u(i,:);
        vi = v(i,:);
        
        % Form right side of algebraic equation
        bet = omi;
        if i==id
            bet = bet - (dt/2/rho)*dfx';
        end
        pidx = find(ui>=0);
        bet(pidx) = bet(pidx)-(dt/2/dx)*(uom(i,pidx)-uom(i-1,pidx));
        nidx = find(ui<0);
        bet(nidx) = bet(nidx)-(dt/2/dx)*(uom(i+1,nidx)-uom(i,nidx));
        bet([1 end]) = 0;
        bet = bet(:);
        
        % Form tridiagonal matrix of coefficients based on upwinding
        c = -(dt/2/dr)*( vi(2:end-1)>=0 ).*vi(1:end-2);
        c = [c(:); 0; 0];
        b = 1+(dt/2/dr)*( vi(2:end-1)>=0 ).*vi(2:end-1);
        b = b -  (dt/2/dr)*( vi(2:end-1)<0 ).*vi(2:end-1);
        b = [1; b(:); 1];
        a = (dt/2/dr)*( vi(2:end-1)<0 ).*vi(3:end);
        a = [0; 0; a(:)];
        M = spdiags([c b a],[-1 0 1],Nr,Nr);
        
        % Solve equations for updated vorticity
        om(i,:) = M\bet;
    end
    om(end,:) = om(end-1,:);
    
    % Solve Poisson Equation: Update streamfunction
    Bi = rmat.*om;
    Bi = Bi(2:Nx-1,2:Nr-1);
    Bp = [Bi(:); Bbc(:)];
    
    %psi = Mp\Bp;
    psi = Up\(Lp\Bp);
    psi = reshape(psi,[Nx,Nr]);
    
    % Update velocity field
    u = repmat(U0,[Nx Nr]);
    u(:,2:Nr-1) = -( psi(:,3:Nr)-psi(:,1:Nr-2) )./rmat(:,2:Nr-1)/2/dr;
    u(:,1) = 4/3*u(:,2)-1/3*u(:,3);
        
    v = zeros(Nx,Nr);
    v(2:Nx-1,:) = ( psi(3:Nx,:) - psi(1:Nx-2,:) )./rmat(2:Nx-1,:)/2/dx;
    v(1,:) = ( -3*psi(1,:) + 4*psi(2,:)-psi(3,:) )./rmat(1,:)/2/dx;
    v(end,:) = ( 3*psi(end,:) - 4*psi(end-1,:)+psi(end-2,:) )./rmat(end,:)/2/dx;
    v(:,1) = 0;
end

%% Steady-state operating characteristics

% Compute average rotor plane velocity Ud, m/s
Ud = ( u(id,1:jd)*r(1:jd)*2*pi*dr ) / (pi*R^2);
fprintf('\nAverage rotor plane velocity Ud = %4.3f m/s\n',Ud);

% Compute induction factor
a = 1-Ud/U0;
fprintf('Induction factor a = %4.3f\n',a);

% Compute total axial force and compare with Betz theory
Fx = fx*dx;
T = abs(Fx(1:jd)'*r(1:jd)*2*pi*dr);
%[CT T/(0.5*rho*U0^2*Ad) 4*a*(1-a)];
CTSim = T/(0.5*rho*U0^2*Ad);
CTBetz = 4*a*(1-a);
fprintf('Total axial force T = %4.3fN\n',T);
fprintf('CT(Chosen) = %4.4f \t CT(Sim) = %4.4f, CT(Betz) = %4.4f\n',...
    CT,CTSim,CTBetz);

% Compute generated power and compare with Betz theory
Pgen = sum( -Fx(1:jd).*u(id,1:jd)'.*r(1:jd)*(2*pi*dr) );
Pwind = 0.5*rho*U0^3*Ad;
CPSim = Pgen / Pwind;
CPBetz = 4*a*(1-a)^2;

fprintf('Generated Power Pgen = %4.3fkW\n',Pgen/1e3);
fprintf('CP(Sim) = %4.4f, CP(Betz) = %4.4f\n',CPSim,CPBetz);

%return
% Mesh of streamwise and radial velocity distributions
figure(1)
mesh(x,r,u');
xlabel('x, m');
ylabel('r, m');
zlabel('u, m/s');

figure(2)
mesh(x,r,v');
xlabel('x, m');
ylabel('r, m');
zlabel('v, m/s');

figure(3)
mesh(x,r,psi');
xlabel('x, m');
ylabel('r, m');
zlabel('psi, m^3/s');

